/*

(c); Copyright, Real-Time Innovations, 2023.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_DIST_LOGGER_HPP_
#define RTI_DIST_LOGGER_HPP_

#include <rti/distlogger/detail/macros.hpp>
#include <rti/distlogger/DistLoggerOptions.hpp>
#include <rti/distlogger/detail/DistLoggerImpl.hpp>
#include <dds/core/Reference.hpp>

namespace rti { namespace dist_logger {

/** 
 *  @defgroup RTI_DL_DistLogger_Module Modern C++ RTI Distributed Logger Classes and API
 *  @brief Main APIs to use the Modern C++ RTI Distributed Logger.
*/

/**
 *  @class DistLogger
 *  @ingroup RTI_DL_DistLogger_Module
 *
 *  @brief This class contains RTI Distributed Logger singleton and the methods
 *  to configure and use it.
 *
 *  This class is a <b>singleton</b>, so only one copy will be created.
 *  Use the \ref DistLoggerOptions class to customize the singleton \b BEFORE
 *  creating it.
 */
class RTIDLCPP2DllExport DistLogger : public dds::core::Reference<DistLoggerImpl> {
public:

    OMG_DDS_REF_TYPE_NODC(DistLogger, dds::core::Reference, DistLoggerImpl);

    /**
     *  @brief Provides access to the DistLogger singleton.
     *
     *  @details This method provides access to the singleton instance of the DistLogger. If
     *  the instance does not exist, it will be created. Note that no change in
     *  options is allowed after the singleton is created.
     *  @return The singleton instance of the DistLogger.
     *  @throws dds::core::Error on failure.
     */
    static DistLogger& get_instance();

    /**
     *  @brief Finalizes the DistLogger singleton.
     *
     *  @details This method must be called before the application ends.
     *  @throws dds::core::Exception on failure.
     */
    static void finalize();

    /**
     *  @brief Assigns configurable options to RTI Distributed Logger before its
     *  creation.
     *
     *  @details Assigns the values in the \c options to the \ref DistLogger
     *  singleton \b before its creation. If the singleton was already created,
     *  this method has no effect.
     *  If a Domain Participant is provided ih the \c options, the Distributed Logger
     *  instance must be finalized before finalizing the Domain Participant Factory.
     *
     *  @param options The new options to be assigned to the \ref DistLogger singleton.
     *  @throws dds::core::IllegalOperationError when the \ref DistLogger singleton has
     *  already been created or dds::core::Exception on failure setting the options.
     */
    static void set_options(const rti::dist_logger::DistLoggerOptions& options);

    /**
     *  @brief Sets the filter level for the singleton.
     *
     *  @details Filtering controls what type of log messages are sent.
     *  This enables filtering on the \em sending side (in the application using 
     *  Distributed Logger). You can also enable filtering on the \em receiving
     *  side (such as in a tool that receives log messages, like RTI Admin
     *  Console or RTI Monitor).
     *
     *  Supported Filter Levels: See the table below.
     *  The default value is RTI_DL_INFO_LEVEL.
     *  The default value is RTI_DL_INFO_LEVEL.
     *
     *  Or see the IDL file provided with Distributed Logger
     *  (file: \ref_url_dl_idl).In the IDL file, there are language-specific bindings.
     *  Prepend the appropriate binding to the filter level shown in the IDL. 
     *  For example, one of the filter levels in the IDL is INFO_LEVEL. 
     *  To refer to this filter level in C++, use RTI_DL_INFO_LEVEL.
     *
     *  Log messages up to and including the specified filter level will pass
     *  through the filter and be sent on the network. For example, the WARNING
     *  level allows log messages up to and including the WARNING level (FATAL,
     *  SEVERE, ERROR, and WARNING) to be sent; any other log messages with
     *  numerically higher filter levels (INFO, DEBUG, and TRACE) will be
     *  filtered out and not sent. 
     *
     *  @param new_level The new filter level.
     *  @throws dds::core::Exception on failure.
     *
     *  The following table shows how RTI Connext DDS log levels map to RTI
     *  Distributed Logger log filter levels.
     *
     *  Not all RTI Distributed Logger levels appear in this table because there
     *  is not a one-to-one mapping between RTI Connext DDS and RTI Distributed
     *  Logger levels (there are more levels in RTI Distributed Logger).
     *  You can obtain the filter level using a constant defined in the Distributed
     *  Logger IDL file or a getter function in the Distributed Logger API.
     *
     * <CENTER><TABLE>
     * <CAPTION><EM>Filter Levels</EM></CAPTION>
     * <TR>
     * <TH> Connext DDS API constant </TH>
     * <TH> Distributed Logger IDL constant </TH>
     * <TH> Distributed Logger API getter </TH>
     * </TR>
     * <TR>
     *    <TD>rti::config::LogLevel::FATAL_ERROR</TD>
     *    <TD>RTI_DL_FATAL_LEVEL</TD>
     *    <TD>\ref get_fatal_log_level</TD>
     * </TR>
     * <TR>
     *    <TD>rti::config::LogLevel::EXCEPTION</TD>
     *    <TD>RTI_DL_ERROR_LEVEL</TD>
     *    <TD>\ref get_error_log_level</TD>
     * </TR>
     * <TR>
     *    <TD>rti::config::LogLevel::WARNING</TD>
     *    <TD>RTI_DL_WARNING_LEVEL</TD>
     *    <TD>\ref get_warning_log_level</TD>
     * </TR>
     * <TR>
     *    <TD>rti::config::LogLevel::STATUS_LOCAL</TD>
     *    <TD>RTI_DL_NOTICE_LEVEL</TD>
     *    <TD>\ref get_notice_log_level</TD>
     * </TR>
     * <TR>
     *    <TD>rti::config::LogLevel::STATUS_REMOTE</TD>
     *    <TD>RTI_DL_INFO_LEVEL</TD>
     *    <TD>\ref get_info_log_level</TD>
     * </TR>
     * <TR>
     *    <TD>rti::config::LogLevel::STATUS_ALL</TD>
     *    <TD>RTI_DL_DEBUG_LEVEL</TD>
     *    <TD>\ref get_debug_log_level</TD>
     * </TR>
     * </TABLE></CENTER>
     *
     */
    void set_filter_level(int new_level);

    /**
     *  @brief Sets PrintFormat of RTI Logger to the provided value.
     *
     *  @details This method is mostly a pass-through to the appropriate call in RTI
     *  DDS library, but it also has the benefit of keeping the state of RTI
     *  Logger and RTI Distributed Logger in sync.
     *  @param log_format The new PrintFormat.
     *  @throws dds::core::Error on failure.
     */
    void set_print_format(const rti::config::PrintFormat& log_format);

    /**
     *  @brief Updates the rti::config::LogCategory and the
     *  rti::config::Verbosity in RTI Logger to the provided value.
     *
     *  @details This method is mostly a pass-through to the appropriate call in RTI
     *  DDS library, but it also has the benefit of keeping the state of RTI
     *  Logger and RTI Distributed Logger in sync.
     *  @param category The category for which to set the verbosity.
     *  @param verbosity The new verbosity level for the category.
     *  @throws dds::core::Error on failure.
     */
    void set_verbosity(const rti::config::LogCategory& category, const rti::config::Verbosity& verbosity);

    /**
     *  @brief Logs a message at the specified level and with the specified
     *  category.
     *
     *  @param log_level The level of the log message.
     *  @param message The message to log.
     *  @param category The category of the log message.
     */
    void log_message(int log_level, const char * message, const char * category);

    /**
     *  @brief Logs a message at the specified level and with the specified
     *  category.
     *
     *  @param log_level The level of the log message.
     *  @param message The message to log.
     *  @param category The category of the log message.
     */
    void log_message(int log_level, const std::string& message, const std::string& category);
   
    /**
     *  @brief Performs the same function as
     *  \ref log_message(int, const std::string&, const std::string&)
     *  except that it also allows specification of the message timestamp.
     *
     *  @param log_level The level of the log message.
     *  @param message The message to log.
     *  @param category The category of the log message.
     *  @param timestamp The timestamp of the log message.
     */
    void log_message(int log_level, const char * message, const char * category, const dds::core::Time& timestamp);

    /**
     *  @brief Performs the same function as
     *  \ref log_message(int, const std::string&, const std::string&)
     *  except that it also allows specification of the message timestamp.
     *
     *  @param log_level The level of the log message.
     *  @param message The message to log.
     *  @param category The category of the log message.
     *  @param timestamp The timestamp of the log message.
     */
    void log_message(int log_level, const std::string& message, const std::string& category, const dds::core::Time& timestamp);

    /**
     *  @brief Logs a fatal message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_FATAL_LEVEL log level.
     *  @param message The message to log.
     */
    void fatal(const char * message);

    /**
     *  @brief Logs a fatal message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_FATAL_LEVEL log level.
     *  @param message The message to log.
     */
    void fatal(const std::string& message);

    /**
     *  @brief Logs a severe message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_SEVERE_LEVEL log level.
     *  @param message The message to log.
     */
    void severe(const char * message);

    /**
     *  @brief Logs a severe message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_SEVERE_LEVEL log level.
     *  @param message The message to log.
     */
    void severe(const std::string& message);
    
    /**
     *  @brief Logs a error message
     *
     *  @details The message received is logged using  the default category and the
     *  RTI_DL_ERROR_LEVEL log level.
     *  @param message The message to log.
     */
    void error(const char * message);

    /**
     *  @brief Logs a error message
     *
     *  @details The message received is logged using  the default category and the
     *  RTI_DL_ERROR_LEVEL log level.
     *  @param message The message to log.
     */
    void error(const std::string& message);

    /**
     *  @brief Logs a warning message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_WARNING_LEVEL log level.
     *  @param message The message to log.
     */
    void warning(const char * message);

    /**
     *  @brief Logs a warning message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_WARNING_LEVEL log level.
     *  @param message The message to log.
     */
    void warning(const std::string& message);

    /**
     *  @brief Logs a notice message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_NOTICE_LEVEL log level.
     *  @param message The message to log.
     */
    void notice(const char * message);

    /**
     *  @brief Logs a notice message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_NOTICE_LEVEL log level.
     *  @param message The message to log.
     */
    void notice(const std::string& message);

    /**
     *  @brief Logs a info message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_INFO_LEVEL log level.
     *  @param message The message to log.
     */
    void info(const char * message);

    /**
     *  @brief Logs a info message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_INFO_LEVEL log level.
     *  @param message The message to log.
     */
    void info(const std::string& message);

    /**
     *  @brief Logs a debug message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_DEBUG_LEVEL log level.
     *  @param message The message to log.
     */
    void debug(const char * message);

    /**
     *  @brief Logs a debug message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_DEBUG_LEVEL log level.
     *  @param message The message to log.
     */
    void debug(const std::string& message);

    /**
     *  @brief Logs a trace message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_TRACE_LEVEL log level.
     *  @param message The message to log.
     */
    void trace(const char * message);

    /**
     *  @brief Logs a trace message
     *
     *  @details The message received is logged using the default category and the
     *  RTI_DL_TRACE_LEVEL log level.
     *  @param message The message to log.
     */
    void trace(const std::string& message);

    /**
     *  @brief Logs a message
     *
     *  @details The message received is logged using the default category and the
     *  given log level.

     *  @param log_level The level of the log message.
     *  @param message The message to log.
     */
    void log(int log_level, const char * message);

    /**
     *  @brief Logs a message
     *
     *  @details The message received is logged using the default category and the
     *  given log level.

     *  @param log_level The level of the log message.
     *  @param message The message to log.
     */
    void log(int log_level, const std::string& message);

    /**
    *  @brief Gets Fatal log level.
    *  @details This method returns the currently defined value for
    *  RTI_DL_FATAL_LEVEL defined in \ref_url_dl_idl
    *  @see \ref set_filter_level
    *  and \ref log_message(int, const std::string&, const std::string&);
    *
    *  @return  Fatal log level.
    */
    int get_fatal_log_level() const;

    /**
    *  @brief Gets Error log level.
    *  @details This method returns the currently defined value for
    *  RTI_DL_ERROR_LEVEL defined in \ref_url_dl_idl
    *  @see \ref set_filter_level
    *  and \ref log_message(int, const std::string&, const std::string&);
    *
    *  @return Error log level.
    */
    int get_error_log_level() const;

    /**
    *  @brief Gets Warning log level.
    *  @details This method returns the currently defined value for
    *  RTI_DL_WARNING_LEVEL defined in \ref_url_dl_idl
    *  @see \ref set_filter_level
    *  and \ref log_message(int, const std::string&, const std::string&);
    *
    *  @return Warning log level.
    */
    int get_warning_log_level() const;

    /**
    *  @brief Gets Notice log level.
    *  @details This method returns the currently defined value for
    *  RTI_DL_NOTICE_LEVEL defined in \ref_url_dl_idl
    *  @see \ref set_filter_level
    *  and \ref log_message(int, const std::string&, const std::string&);
    *
    *  @return Notice log level.
    */
    int get_notice_log_level() const;

    /**
    *  @brief Gets Info log level.
    *  @details This method returns the currently defined value for
    *  RTI_DL_INFO_LEVEL defined in \ref_url_dl_idl
    *  @see \ref set_filter_level
    *  and \ref log_message(int, const std::string&, const std::string&);
    *
    *  @return Info log level.
    */
    int get_info_log_level() const;

    /**
    *  @brief Gets Debug log level.
    *  @details This method returns the currently defined value for
    *  RTI_DL_DEBUG_LEVEL defined in \ref_url_dl_idl
    *  @see \ref set_filter_level
    *  and \ref log_message(int, const std::string&, const std::string&);
    *
    *  @return Debug log level.
    */
    int get_debug_log_level() const;

private:
    DistLogger();

    static DistLogger instance_;
};

} } // namespace rti::dist_logger

#endif // RTI_DIST_LOGGER_HPP_
