/*

(c) Copyright, Real-Time Innovations, 2023.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_DIST_LOGGER_OPTIONS_HPP_
#define RTI_DIST_LOGGER_OPTIONS_HPP_

#include <rti/distlogger/detail/macros.hpp>
#include <rti/core/ThreadSettings.hpp>
#include <dds/domain/DomainParticipant.hpp>


namespace rti { namespace dist_logger {

/** 
 *  @class DistLoggerOptions
 *  @relates DistLogger
 *  @ingroup RTI_DL_DistLogger_Module
 *  @headerfile rti/distlogger/DistLoggerOptions.hpp
 *  @brief This class contains RTI Distributed Logger configurable options.
 *
 *  @details Use this API to customize RTI Distributed Logger to adapt it to your needs.
 *  Assign the newly created DistLoggerOptions object to the RTI Distributed Logger
 *  singleton before calling to DistLogger::get_instance().
 */
class RTIDLCPP2DllExport DistLoggerOptions {
public:

    /**
     *  @brief Creates a new object.
     */
    DistLoggerOptions();

    /** 
     *  @brief Gets currently set Domain Participant.
     *
     *  @return The current Domain Participant if set. Otherwise, dds::core::null.
     */
    dds::domain::DomainParticipant domain_participant() const;

    /** 
     *  @brief Sets current Domain Participant.
     *
     *  @details There are several use cases in which you may want to use your own
     *  Domain Participant to create RTI Distributed Logger internal DDS
     *  entities. By calling this function, you will have control of the
     *  Domain Participant used by RTI Distributed Logger. Once assigned,
     *  \p domain_participant will be used to create the DDS
     *  entities (Topics, Publishers, Subscribers, DataWriters, & DataReaders)
     *  needed by RTI Distributed Logger.
     *  If \p domain_participant is set, the associated Distributed Logger instance
     *  must be finalized before finalizing the Domain Participant Factory.
     *
     *  @param domain_participant The dds::domain::DomainParticipant to use.
     */
    void domain_participant(const dds::domain::DomainParticipant& domain_participant);

    /**
     *  @brief Indicates whether or not remote administration is enabled.
     *
     *  @details If true, remote administration Topics will be created and
     *  used. If false, they will not be created. The default value
     *  is true.
     *  @return true if remote administration is enabled. Otherwise false.
     */
    bool remote_administration_enabled() const;

    /**
     *  @brief Enables or disables remote administration.
     *
     *  @details Sets remote administration of the Topics to the value passed as
     *  parameter.
     *  @param enabled A bool indicating whether or not to enable remote administration.
     */
    void remote_administration_enabled(bool enabled);

    /** 
     *  @brief Gets the log filter level.
     *
     *  @details Supported Filter Levels: See the IDL file provided with RTI Distributed
     *  Logger (\ref_url_dl_idl). The default value is RTI_DL_INFO_LEVEL.
     *
     *  @return integer containing the log filter level of the options.
     */
    int filter_level() const;

    /** 
     *  @brief Sets the log filter level.
     *
     *  @details Filtering controls what type of log messages are sent.
     *  This enables filtering on the \em sending side (in the application using
     *  RTI Distributed Logger). You can also enable filtering on the
     *  \em receiving side (such as in a tool that receives log messages, like RTI
     *  Admin Console or RTI Monitor).
     *
     *  Supported Filter Levels: See the IDL file provided with RTI Distributed
     *  Logger (\ref_url_dl_idl). The default value is RTI_DL_INFO_LEVEL.
     *
     *  @param filter_level The log filter level of the options (see above).
    */
    void filter_level(int filter_level);

    /**
     *  @brief Gets the current Domain ID.
     *
     *  Returns the Domain ID to be used by RTI Distributed Logger. This value is
     *  only used if a Domain Participant has not been set. Otherwise,
     *  the Domain ID used is the one of the set Domain Participant
     *
     *  The default value is 0.
     *  @return Integer containing the DDS Domain ID where RTI Distributed
     *  Logger instance will be created.
     */
    int domain_id() const;

    /** 
     *  @brief Sets the current Domain ID
     *
     *  Update the Domain ID to be used by RTI Distributed Logger. This value is
     *  only used if a Domain Participant has not been set. Otherwise,
     *  the Domain ID used is the one of the set Domain Participant.
     *
     *  The default value is 0.
     *  @param domain_id Int representing a DDS Domain ID. Max value: 250
     */    
    void domain_id(int domain_id);

    /**
     *  @brief Gets the Qos Library that will be used to create to RTI
     *  Distributed Logger
     *
     *  @details The QoS library name to use when creating DDS entities. It is up to the
     *  application to ensure that this library is available at run time for the 
     *  Distributed Logger to use.
     *  @return rti::core::optional_value<std::string> containing the current
     *  Qos Library.
     */
    rti::core::optional_value<std::string> qos_library() const;

    /**
     *  @brief Sets the Qos Library that will be used to create to RTI
     *  Distributed Logger
     *
     *  @details The QoS Library name to use when creating DDS entities. It is up to the
     *  application to ensure that this library is available at run time for the 
     *  Distributed Logger to use.
     *  @param qos_library rti::core::optional_value<std::string> containing the Qos Library
     *  to set.
     */
    void qos_library(const rti::core::optional_value<std::string>& qos_library);

    /**
     *  @brief Gets the Qos Profile that will be used to create to RTI
     *  Distributed Logger
     *
     *  @details The QoS Profile name to use when creating DDS entities. It is up to the
     *  application to ensure that this profile is available at run time for the 
     *  Distributed Logger to use.
     *  @return rti::core::optional_value<std::string> containing the current
     *  Qos Profile.
     */
    rti::core::optional_value<std::string> qos_profile() const;

    /**
     *  @brief Sets the Qos Profile that will be used to create to RTI
     *  Distributed Logger
     *
     *  @details The QoS Profile name to use when creating DDS entities. It is up to the
     *  application to ensure that this profile is available at run time for the 
     *  Distributed Logger to use.
     *  @param qos_profile rti::core::optional_value<std::string> containing the Qos Profile
     *  to set.
     */
    void qos_profile(const rti::core::optional_value<std::string>& qos_profile);

   /**
     *  @brief Gets the maximum queue size that will be assigned to RTI
     *  Distributed Logger
     *
     *  @details This field controls the size of the queue used to temporarily store log
     *  messages from the application until they can be written out to DDS. The default
     *  value is RTI_DL_DEFAULT_QUEUE_SIZE.
     *  @return Integer containing the current maximum queue size.
    */
    int queue_size() const;

   /**
     *  @brief Sets the maximum queue size that will be assigned to RTI
     *  Distributed Logger
     *
     *  @details This field controls the size of the queue used to temporarily
     *  store log messages from the application until they can be written out to
     *  DDS.
     *  @param queue_size Integer containing the new maximum queue size.
     */
    void queue_size(int queue_size);

    /**
     *  @brief Get the thread settings of the publishing thread that will be
     *  created by RTI Distributed Logger.
     *
     *  @details This field contains general settings of the publishing thread in 
     *  Distributed Logger, like the thread priority, stack size, and other
     *  settings.
     *  @return The currently set rti::core::ThreadSettings
     */
    rti::core::ThreadSettings thread_settings() const;

    /**
     *  @brief Sets the settings of the publishing thread that will be created
     *  by RTI Distributed Logger
     *
     *  @details This parameter is used to control thread parameters such as priority, stack
     *  size, and other settings. These settings will only affect the
     *  publishing thread created by RTI Distributed Logger.
     *  @param thread_settings A valid rti::core::ThreadSettings object with the
     *  settings for Distributed Logger's write thread.
     */
    void thread_settings(const rti::core::ThreadSettings& thread_settings);

    /**
     *  @brief Gets the application kind that will be associated with RTI
     *  Distributed Logger
     *
     *  @details This char pointer is sent along on the state Topic and can help identify
     *  different applications within a system. You can use this field however
     *  you think it will help you most.
     *
     *  @return rti::core::optional_value<std::string> containing the current
     *  application kind.
     */
    rti::core::optional_value<std::string> application_kind() const;

    /**
     *  @brief Sets the application kind that will be associated with RTI
     *  Distributed Logger
     *
     *  @details This string is sent along on the state Topic and can help identify
     *  different applications within a system. You can use this field however
     *  you think it will help you most.
     *  @param application_kind rti::core::optional_value<std::string> containing
     *  the new application kind.
     */
    void application_kind(const rti::core::optional_value<std::string>& application_kind);

    /** 
     *  @brief Indicates whether this feature is enabled.
     *
     *  @details If true, Distributed Logger prints the log messages received to the
     *  stdout. The default value is true.
     *  @return true if echo_to_stdout is enabled. Otherwise false.
     */
    bool echo_to_stdout() const;

    /**
     *  @brief Enables or disables this feature.
     *
     *  @details If true, Distributed Logger prints the log messages received to the 
     *  stdout. The default value is true.
     *  @param echo_to_stdout A bool indicating whether or not to enable this feature.
     */
    void echo_to_stdout(bool echo_to_stdout);
    
    /** 
     *  @brief Indicates whether Distributed Logger should also log all
     *  messages produced by infrastructure code automatically.
     *
     *  @details If true, messages produced by either the middleware or the tools will
     *  be forwarded to Distributed Logger.
     *  @return true if log_infrastructure_messages is enabled. Otherwise false.
     */
    bool log_infrastructure_messages() const;

    /**
     *  @brief Enables or disables automatic logging of infrastructure messages
     *
     *  @details If true, messages produced by either the middleware or the tools will
     *  be forwarded to Distributed Logger. In this mode Distributed Logger uses
     *  NDDS_Config_Logger_set_output_device() API to receive messages from the
     *  middleware. Hence any device previously registered with the
     *  rti::config::Logger::set_device() API will be unregistered.
     *  Similarly Distributed Logger will stop logging infrastructure messages if the
     *  set_device() API is used to register a new device after the
     *  creation of the Distributed Logger instance.
     *
     *  @param log  bool indicating whether or not to enable this feature.
     */
    void log_infrastructure_messages(bool log);

    /**
     *  @brief Compares two objects for equality.
     *
     *  @param other The other object to compare.
     */
    bool operator==(const DistLoggerOptions& other) const;

    /**
     *  @brief Compares two objects for inequality.
     *
     *  @param other The other object to compare.
     */
    bool operator!=(const DistLoggerOptions& other) const;

    /**
     *  @brief Swaps two objects.
     *
     *  @param left One of the objects to swap.
     *  @param right The other object to swap.
     */
    friend void swap(DistLoggerOptions& left, DistLoggerOptions& right) noexcept;

private:
    bool echo_to_stdout_;
    bool log_infrastructure_messages_;
    bool remote_administration_enabled_;
    int queue_size_;
    int domain_id_;
    int filter_level_;
    rti::core::ThreadSettings thread_settings_;
    rti::core::optional_value<std::string> application_kind_;
    rti::core::optional_value<std::string> qos_profile_;
    rti::core::optional_value<std::string> qos_library_;
    dds::domain::DomainParticipant participant_;
};

} } // namespace rti::dist_logger

#endif // RTI_DIST_LOGGER_OPTIONS_HPP_
