/* $Id$

(c) Copyright, Real-Time Innovations, 2018.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.

============================================================================= */

#ifndef RTI_DDS_TOPIC_FLATDATATRAITS_HPP_
#define RTI_DDS_TOPIC_FLATDATATRAITS_HPP_

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

// Include the FlatData API, which is shared among the traditional C++ and
// this modern C++ API
#include "rti/flat/rtiflat.hpp"

#include <type_traits>
#include <dds/core/types.hpp>
#include <rti/topic/cdr/InterpreterHelpers.hpp>

namespace rti { namespace flat { namespace topic {
// is_flat_data_type:
// type trait that indicates if a type is a flat-data type

// General case (false_type)
template <typename TopicType>
struct is_flat_data_type_only : dds::core::false_type {};

template <typename TopicType>
struct is_flat_data_type : dds::core::false_type {};

// Only Sample<OffsetType> types are flat-data types (true_type)
template <typename OffsetType>
struct is_flat_data_type<rti::flat::Sample<OffsetType> > : dds::core::true_type {};

} } } // namespace rti::flat::topic

namespace rti { namespace xcdr {

// Specialization of serialization functions from xcdr for safe_enum, which is
// defined in dds_cpp.2.0
template <typename Enum>
struct primitive_type_traits<dds::core::safe_enum<Enum> > {
    typedef dds::core::safe_enum<Enum> SafeEnum;

    static void serialize_fast(RTIXCdrStream *stream, const SafeEnum* v)
    {
        RTIXCdrStream_serialize4ByteFast(stream, (RTIXCdr4Byte *) v);
    }
    static void deserialize_fast(RTIXCdrStream *stream, SafeEnum* v)
    {
        RTIXCdrStream_deserialize4ByteFast(stream, (RTIXCdr4Byte *) v);
    }
    static bool serialize(RTIXCdrStream *stream, const SafeEnum* v)
    {
        return RTIXCdrStream_serialize4Byte(
                stream, 
                (RTIXCdr4Byte *) v, 
                RTI_XCDR_TRUE) == RTI_XCDR_TRUE;
    }
    static bool deserialize(RTIXCdrStream *stream, SafeEnum* v)
    {
        return RTIXCdrStream_deserialize4Byte(
                stream, 
                (RTIXCdr4Byte *) v, 
                RTI_XCDR_TRUE) == RTI_XCDR_TRUE;
    }
};

} // namespace xcdr

namespace flat { namespace detail {

// This tells the FlatData infrastructure how to default-initialize an enum
// class using the generated default_enumerator traits.
template <typename Enum>
struct default_primitive_value<
        Enum,
        typename std::enable_if<std::is_enum<Enum>::value>::type> {
    static Enum get()
    {
        return rti::topic::default_enumerator<Enum>::value;
    }
};

template <typename InnerEnum, unsigned int N>
struct final_offset_initializer<
    rti::flat::PrimitiveArrayOffset<dds::core::safe_enum<InnerEnum>, N> > {

    static bool initialize(
            rti::flat::PrimitiveArrayOffset<dds::core::safe_enum<InnerEnum>, N>& array)
    {
        dds::core::safe_enum<InnerEnum> default_value = InnerEnum::get_default();
        if (static_cast<int>(default_value.underlying()) == 0) {
            memset(array.get_buffer(), 0, array.get_buffer_size());
        } else {
            for (unsigned int i = 0; i < array.element_count(); i++) {
                array.set_element(i, default_value);
            }
        }
        return true;
    }
};

// indicates that enums require DHeader; defined here because safe_enum is
// a type defined in the Connext C++ headers, not in the Flat Data headers.

template <typename EnumType>
struct primitive_sequence_dheader_gen<
        EnumType,
        typename std::enable_if<std::is_enum<EnumType>::value>::type> {
    using type = DHeaderGeneratorForEnum;
};

template <typename EnumType>
struct primitive_sequence_header_length<
        EnumType,
        typename std::enable_if<std::is_enum<EnumType>::value>::type> {
    static size_t value()
    {
        // The length of a primitive sequence header may be different for enums
        // depending on whether enums are treated as complex. For all primitive
        // types, sequences only contain the sequence lenght (number of
        // elements), but for enums, if they are treated as complex, the header
        // also contains the DHeader (total size of the sequence).
        return sizeof(rti::xcdr::length_t)
                + (is_dheader_required_in_enum_collections()
                           ? RTI_XCDR_DHEADER_SIZE
                           : 0);
    }
};

template <typename EnumType, unsigned int N>
struct is_fixed_type_w_dheader<
        rti::flat::PrimitiveArrayOffset<EnumType, N>,
        typename std::enable_if<std::is_enum<EnumType>::value>::type> {
    static bool value()
    {
        return is_dheader_required_in_enum_collections();
    }
};

} } // namespace flat::detail

namespace topic {

// All flat-data types are non-copyable. By defining this trait as true_type
// an static_assert in DataReader methods that copy the data will fail to 
// compile
template <typename OffsetType>
struct topic_type_disabled_copy<rti::flat::Sample<OffsetType> > :
        dds::core::true_type
{
};

} // namespace topic

} // namespace rti

#endif // RTI_DDS_TOPIC_FLATDATATRAITS_HPP_
