/*
 * (c) Copyright, Real-Time Innovations, 2002-2022.
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 */

#ifndef rtitest_log_impl_h
#define rtitest_log_impl_h

#ifdef __cplusplus
extern "C" {
#endif

/* ======================= RTI TEST LOG API ======================= */

/** @internal
 * @ingroup RTITestLogging
 * @brief Log a message with a custom function name using the the internal
 * logging infrastructure. Expects the log message as a format string and its
 * parameters as variadic arguments, identical to the parameters for printf.
 * This logging infrastructure works similarly to a printf and will output the
 * log message to stdout.
 * @param instrumentationBit__ \b In. RTITestLogBitmap to define the message's
 * verbosity.
 * @param functionName__ \b In. Name of the function that is logging this
 * message.
 * @param logMessage__ \b In. RTITestLogMessage template to define the context
 * of the logging message.
 */
#define RTITestLogger_logWithFunctionName( \
        instrumentationBit__, \
        functionName__, \
        logMessage__, \
        ...) \
  if (RTITestLog_g_instrumentationMask & (instrumentationBit__)) { \
    RTITestLogger_logMessage((functionName__), (logMessage__), __VA_ARGS__); \
  }

/**
 * @ingroup RTITestLogging
 * @brief Log a message using Unit Test Framework internal logging
 * infrastructure. Expects the log message as a format string and its parameters
 * as variadic arguments, identical to the parameters for printf. The Unit Test
 * Framework logging infrastructure works similar to a printf and will output
 * the log message to stdout.
 * @details For example:
 * @code
 * RTITestLogger_log(
 *         RTI_TEST_LOG_BIT_EXCEPTION,
 *         "Testing that %d writers can discover %d readers for topic %s",
 *         numWriters,
 *         numReaders,
 *         topicName);
 * @endcode
 * @param instrumentationBit__ \b In. RTITestLogBitmap to define the message's
 * verbosity.
 * @param logMessage__ \b In. RTITestLogMessage template to define the context
 * of the logging message.
 */
#define RTITestLogger_log(instrumentationBit__, logMessage__, ...) \
  RTITestLogger_logWithFunctionName( \
          (instrumentationBit__), \
          RTI_FUNCTION_NAME, \
          (logMessage__), \
          __VA_ARGS__)

/** @internal
 * @ingroup RTITestLogging
 * @brief Log a message with params using the internal logging
 * infrastructure. Expects the log message as a format string and its parameters
 * as variadic arguments, identical to the parameters for printf. This
 * logging infrastructure works similar to a printf and will output the log
 * message to stdout.
 * @param instrumentationBit__ \b In. RTITestLogBitmap to define the message's
 * verbosity.
 */
#define RTITestLog_logParamString(instrumentationBit__, ...) \
  RTITestLogger_logWithFunctionName( \
          (instrumentationBit__), \
          "", \
          &RTI_TEST_LOG_EMPTY_TEMPLATE, \
          __VA_ARGS__)

/** @internal
 * @ingroup RTITestLogging
 * @brief Log a plain message, ignoring verbosity and without any other
 * information than the log message. Expects log message as a formatted string
 * as variadic arguments, identical to the parameters for printf. This
 * logging infrastructure works similar to a printf and will output the log
 * message to stdout.
 */
#define RTITestLog_printPlain(...) printf(__VA_ARGS__)

/**
 * @defgroup RTITestLogExpectedLoggingClass Expected Log Management
 * @ingroup RTITestLogging
 * @brief A set of macros used to surround expected logging 
 * coming from the tested code. 
 */

/** @internal
 * @ingroup RTITestLogExpectedLoggingClass
 * Internal usage, do not use this macro directly. Use the verbosity-specific
 * versions below. Use the verbosity of the highest verbosity message that you
 * expect to be printed out between the start and end. i.e. If there will be
 * warnings and exceptions printed out, then use the exceptions macro. If there
 * will only be warnings, use the warning one.
 */
#define RTITestLog_printExpectedMessagesStart( \
        expectedPreconditions__, \
        logBit__, \
        msg__) \
  RTITestLog_logParamString((logBit__), "%s\n", (msg__)); \
  if ((expectedPreconditions__)) { \
    RTITestConnext_cachePreconditionDetection(); \
    RTITestConnext_setPreconditionDetection(RTI_FALSE); \
  }

/** @internal
 * @ingroup RTITestLogExpectedLoggingClass
 * Internal usage, do not use this macro directly. Use the verbosity-specific
 * versions below. Use the verbosity of the highest verbosity message that you
 * expect to be printed out between the start and end. i.e. If there will be
 * warnings and exceptions printed out, then use the exceptions macro. If there
 * will only be warnings, use the warning one.
 */
#define RTITestLog_printExpectedMessagesEnd( \
        expectedPreconditions__, \
        logBit__, \
        msg__) \
  if ((expectedPreconditions__)) { \
    RTITestConnext_setPreconditionDetection( \
            RTITestConnext_getPreconditionDetectionCache()); \
  } \
  RTITestLog_logParamString((logBit__), "%s\n", (msg__))

/* ************ Mark the beginning/end of expected error messages. ************/

#define RTI_TEST_LOG_EXPECTED_WARNINGS_START_MSG \
  ("---- expected warnings start ----")
#define RTI_TEST_LOG_EXPECTED_WARNINGS_END_MSG \
  ("---- expected warnings end ----")

#define RTI_TEST_LOG_EXPECTED_ERRORS_START_MSG \
  ("---- expected errors start ----")
#define RTI_TEST_LOG_EXPECTED_ERRORS_END_MSG ("---- expected errors end ----")

/** @internal
 * @ingroup RTITestLogging
 * @brief Starts a expected error area. This macro is not thread safe when
 * passing true because a global variable is modified.
 * @param expectedPreconditions__ \b In. RTI_TRUE if any precondition are
 * expected to fail, RTI_FALSE otherwise.
 */
#define RTITestLog_printExpectedErrorsStartEx(expectedPreconditions__) \
  RTITestLog_printExpectedMessagesStart( \
          (expectedPreconditions__), \
          RTI_TEST_LOG_BIT_EXCEPTION, \
          RTI_TEST_LOG_EXPECTED_ERRORS_START_MSG)

/**
 * @ingroup RTITestLogExpectedLoggingClass
 * @brief Starts a expected errors area, preconditions are not expected to fail.
 * This macro is not thread safe when passing true because a global variable is
 * modified.
 */
#define RTITestLog_printExpectedErrorsStart() \
  RTITestLog_printExpectedErrorsStartEx(RTI_FALSE)

/** @internal
 * @ingroup RTITestLogging
 * @brief Finishes a expected errors area. This macro is not thread safe when
 * passing true because a global variable is modified.
 * @param expectedPreconditions__ \b In. RTI_TRUE if any precondition are
 * expected to fail, RTI_FALSE otherwise.
 */
#define RTITestLog_printExpectedErrorsEndEx(expectedPreconditions__) \
  RTITestLog_printExpectedMessagesEnd( \
          (expectedPreconditions__), \
          RTI_TEST_LOG_BIT_EXCEPTION, \
          RTI_TEST_LOG_EXPECTED_ERRORS_END_MSG)

/**
 * @ingroup RTITestLogExpectedLoggingClass
 * @brief Finishes a expected errors area, preconditions are not expected to
 * fail. This macro is not thread safe when passing true because a global
 * variable is modified.
 */
#define RTITestLog_printExpectedErrorsEnd() \
  RTITestLog_printExpectedErrorsEndEx(RTI_FALSE)

/* ****************************************************************************/
/* *********** Mark the beginning/end of expected warning messages. ***********/
/** @internal
 * @ingroup RTITestLogging
 * @brief Starts a expected warnings area. This macro is not thread safe when
 * passing true because a global variable is modified.
 * @param expectedPreconditions__ \b In. RTI_TRUE if any precondition are
 * expected to fail, RTI_FALSE otherwise.
 */
#define RTITestLog_printExpectedWarningsStartEx(expectedPreconditions__) \
  RTITestLog_printExpectedMessagesStart( \
          (expectedPreconditions__), \
          RTI_TEST_LOG_BIT_WARN, \
          RTI_TEST_LOG_EXPECTED_WARNINGS_START_MSG)

/**
 * @ingroup RTITestLogExpectedLoggingClass
 * @brief Starts a expected warnings area, preconditions are not expected to
 * fail. This macro is not thread safe when passing true because a global
 * variable is modified.
 */
#define RTITestLog_printExpectedWarningsStart() \
  RTITestLog_printExpectedWarningsStartEx(RTI_FALSE)

/** @internal
 * @ingroup RTITestLogging
 * @brief Finishes a expected warnings area. This macro is not thread safe when
 * passing true because a global variable is modified.
 * @param expectedPreconditions__ \b In. RTI_TRUE if any precondition are
 * expected to fail, RTI_FALSE otherwise.
 */
#define RTITestLog_printExpectedWarningsEndEx(expectedPreconditions__) \
  RTITestLog_printExpectedMessagesEnd( \
          (expectedPreconditions__), \
          RTI_TEST_LOG_BIT_WARN, \
          RTI_TEST_LOG_EXPECTED_WARNINGS_END_MSG)

/**
 * @ingroup RTITestLogExpectedLoggingClass
 * @brief Finishes a expected warnings area, preconditions are not expected to
 * fail. This macro is not thread safe when passing true because a global
 * variable is modified.
 */
#define RTITestLog_printExpectedWarningsEnd() \
  RTITestLog_printExpectedWarningsEndEx(RTI_FALSE)

/* ***************************************************************************/

/** @internal
 * @ingroup RTITestLogging
 * @brief Prints the test name appending extName__ at the end. We need this
 * function because we need all of the test names to be unique, and in some
 * cases that was not the case as we were running the same test but changing the
 * type used by the test. :? is mandatory because RTITestLog_warn added : at the
 * beginning of extName__.
 * @param extName__ \b In. String to be appended just after the test name.
 * @param functionName__ \b In. Function name where the test is located (usually
 * the tester or the submodule test).
 */
#define RTITestLog_printTestNameExtWithFunctionName(functionName__, extName__) \
  RTITestLogger_logWithFunctionName( \
          RTI_TEST_LOG_BIT_EXCEPTION, \
          (functionName__), \
          &RTI_TEST_LOG_EMPTY_TEMPLATE, \
          "%s%s?", \
          (extName__) == NULL ? "" : (extName__), \
          (extName__) == NULL ? "" : ":")

/** @internal
 * @ingroup RTITestLogging
 * @brief Prints the test name appending extName__ at the end. We need this
 * function because we need all of the test names to be unique, and in some
 * cases that was not the case as we were running the same test but changing the
 * type used by the test. :? is mandatory because RTITestLog_warn added : at the
 * beginning of extName__.
 * @param extName__ \b In. String to be appended just after the test name.
 */
#define RTITestLog_printTestNameExt(extName__) \
  RTITest_g_currentTestName = RTI_FUNCTION_NAME; \
  RTITestLog_printTestNameExtWithFunctionName(RTI_FUNCTION_NAME, (extName__))

/**
 * @ingroup RTITestLogging
 * @brief Prints the test name.
 */
#define RTITestLog_printTestName() \
  RTITest_g_currentTestName = RTI_FUNCTION_NAME; \
  RTITestLog_printTestNameExtWithFunctionName(RTI_FUNCTION_NAME, NULL)

/** @internal
 * @ingroup RTITestLogging
 * @brief Prints a log message with the error code explanation of a OS API
 * failure.
 * @param syscallName__ \b In. String to be appended just after the test name.
 */
#define RTITestLog_logOsFailure(syscallName__) \
  RTITestLogger_log( \
          RTI_TEST_LOG_BIT_EXCEPTION, \
          &RTI_TEST_LOG_OS_FAILURE_TEMPLATE, \
          "%s with error code: %d", \
          (syscallName__), \
          RTITestInternal_getError());

/** @internal
 * Internal usage only, these macros are used to get a const string from a
 * parameter's name
 */
#define RTI_TEST_STRINGIFY(x) #x
#define RTI_TEST_TOSTRING(x) RTI_TEST_STRINGIFY(x)

/** @internal
 * @ingroup RTITestLogging
 * @brief Checks whether a particular instrumentation bit is currently set.
 * @param instrumentationBit__ \b In. Instrumentation in question.
 * @return RTI_TRUE if yes, RTI_FALSE otherwise
 */
#define RTITestLog_matchesVerbosity(instrumentationBit__) \
  (RTITestLog_g_instrumentationMask & (instrumentationBit__))

/* ======================== LEGACY LOG API ======================== */

/* Legacy API, use RTITestLog_logParamString instead */
#define RTITestLog_freeForm(instrumentationBit__, ...) \
  RTITestLog_logParamString(instrumentationBit__, __VA_ARGS__)

/* Legacy API, use RTITestLogger_logWithFunctionName instead */
#define RTITestLog_logWithFunctionName( \
        instrumentationBit__, \
        functionName__, \
        ...) \
  if (RTITestLog_g_instrumentationMask & (instrumentationBit__)) \
    RTILogMessage_printWithParams( \
            (RTI_LOG_PRINT_FORMAT_MASK_ALL) \
                    & (RTILogPrintFormatKindMask) ~( \
                            RTI_LOG_PRINT_BIT_LOG_LEVEL), \
            (instrumentationBit__), \
            MODULE_TEST, \
            RTI_FILE_NAME, \
            RTI_LINE_NUMBER, \
            (functionName__), \
            __VA_ARGS__)

/* Legacy API, use RTITestLogger_log instead */
#define RTITestLog_fatal(...) \
  RTITestLog_logWithFunctionName( \
          RTI_TEST_LOG_BIT_FATAL_ERROR, \
          METHOD_NAME, \
          __VA_ARGS__)

/* Legacy API, use RTITestLogger_log instead */
#define RTITestLog_exception(...) \
  RTITestLog_logWithFunctionName( \
          RTI_TEST_LOG_BIT_EXCEPTION, \
          METHOD_NAME, \
          __VA_ARGS__)

/* Legacy API, use RTITestLogger_log instead */
#define RTITestLog_warn(...) \
  RTITestLog_logWithFunctionName( \
          RTI_TEST_LOG_BIT_WARN, \
          METHOD_NAME, \
          __VA_ARGS__)

/* Legacy API, use RTITestLogger_log instead */
#define RTITestLog_local(...) \
  RTITestLog_logWithFunctionName( \
          RTI_TEST_LOG_BIT_LOCAL, \
          METHOD_NAME, \
          __VA_ARGS__)

/* Legacy API, use RTITestLogger_log instead */
#define RTITestLog_remote(...) \
  RTITestLog_logWithFunctionName( \
          RTI_TEST_LOG_BIT_REMOTE, \
          METHOD_NAME, \
          __VA_ARGS__)

/* Legacy API, use RTITestLogger_log instead */
#define RTITestLog_periodic(...) \
  RTITestLog_logWithFunctionName( \
          RTI_TEST_LOG_BIT_PERIODIC, \
          METHOD_NAME, \
          __VA_ARGS__)

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* rtitest_log_impl_h */
